/*
				Copyright 2009 Zhao Lin <linjiework@163.com>
        
        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.
        
        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.
        
        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
        MA 02110-1301, USA.
 */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "global.h"
#include "socket.h"

static DialogImpl *dlg = NULL;
static int  timeout_time = 30;				// sec
static char cookie_string[BUFFERMIDLENGTH];

int socket_connect(const char *server, int port);
int socket_close(int sockFd);
int socket_receive(int sockFd, int length, void *data);
int socket_send(int sockFd, int length, void *data);

static int clear_last_char(char *info);
static int translate_zipdata(const char *buffer, int length, char *data, int len);
static int translate_zipdata_chunked(const char *buffer, int length, char *data, int len);

int socket_init(DialogImpl *pdlg)
{
	dlg = pdlg;
	return 0;
}

int socket_connect(const char *server, int port)
{
	int ret 			= -1;
	socklen_t len = -1;
	int clientfd 	= -1;
	unsigned long ul 				= -1 ;
	unsigned server_ipaddr 	= -1;
	struct sockaddr_in	host_addr;
	struct in_addr 			in;
	struct hostent		 *server_info;
	struct timeval      timeout;
	fd_set              rmask;

	dlg->ShowMessage("Prepare to analyze the server(%s).", server);
	bzero(&host_addr, sizeof(host_addr));
	if ((server_ipaddr = inet_addr(server)) != INADDR_NONE) {
		host_addr.sin_addr.s_addr = server_ipaddr;
	} else
	if ((server_info = gethostbyname(server)) != NULL) {
		bcopy(server_info->h_addr, (char *)&host_addr.sin_addr, server_info->h_length);
	} else {
		dlg->ShowMessage("Can't get the ip of host:%s.", server);
		return -1;
	}

	in.s_addr = host_addr.sin_addr.s_addr;
	// printf("Get host IP: %s\n", inet_ntoa(in));
	dlg->ShowMessage("The IP of server(%s) is %s.", server, inet_ntoa(in));
	dlg->ShowMessage("Prepare to connect the server(%s).", server);
	
	host_addr.sin_family = AF_INET;
	host_addr.sin_port = htons(port);

	clientfd = socket(AF_INET, SOCK_STREAM, 0);
	if (clientfd < 0) {
		dlg->ShowMessage("Create socket error.");
		return -1;
	}
	ul = 1; ioctl(clientfd, FIONBIO, &ul);
	if ((ret = connect(clientfd, (struct sockaddr *)&host_addr, sizeof(host_addr))) == -1) {
		if (errno == EINPROGRESS) {
			timeout.tv_sec  = timeout_time > 6 ? timeout_time : 6;
			timeout.tv_usec = 0;
			FD_ZERO(&rmask);
			FD_SET(clientfd, &rmask);
			if (select(clientfd + 1, NULL, &rmask, NULL, &timeout) > 0) {
				len = sizeof(int);
				getsockopt(clientfd, SOL_SOCKET, SO_ERROR, &ret, &len);
				if (ret != 0)
					ret = -1;
			} else
				ret = -1;
		} else
			ret = -1;
	}
	if (ret != 0) {
		dlg->ShowMessage("Connect to the server(%s:%d, IP:%s) error.", server, port, inet_ntoa(in));
		close(clientfd);
		return -1;
	}
	ul = 0; ioctl(clientfd, FIONBIO, &ul);
	dlg->ShowMessage("Connect to the server(%s:%d, IP:%s) success.", server, port, inet_ntoa(in));

	return clientfd;
}


int socket_close(int sockFd)
{
	int ret = 0;

	if (sockFd > 0) {
		ret = close(sockFd);
	}
	if (ret < 0) {
		return ret;
	} else {
		return 0;
	}
}

static int socket_get_cookie(const char *buffer_start, const char *buffer_end)
{
	const char *p = buffer_start, *p1;
	char  cookie_buffer[BUFFERMIDLENGTH];
	int   cookie_flag = -1;

	if (buffer_start == NULL || buffer_end == NULL) {
		return 0;
	}

	while ((p = strstr(p, "Set-Cookie: ")) != NULL) {
		if (p > buffer_end) {
			// read head finished.
			break;
		}
		if (cookie_flag == -1) {
			// have new cookie, clear old data.
			// infact, we can't to do so, but I ...
			cookie_string[0] 	= 0;
			cookie_flag 			= 0;
		}
		p += strlen("Set-Cookie: ");
		p1 = p;
		while (*p1 != ';' && *p1 != ' ' && *p1 != 0x0A && *p1 != 0x0D && *p1 != 0) {
			p1++;
		}
		strncpy(cookie_buffer, p, p1-p);
		cookie_buffer[p1-p] = 0;
		strcat(cookie_string, cookie_buffer);
		strcat(cookie_string, "; ");
	}
	if (cookie_flag == 0) {
		clear_last_char(cookie_string);
		clear_last_char(cookie_string);
	}
	return 0;
}

// anlyze the chunked web page.
// we operate the data in one string, it's very danger that maybe the data will disordered.
static int socket_analyze_chunked(char *body)
{
	char szSectionLen[64], *pEndPos, *pStartPos, *pCopyPos;
	int  nSectionLen;

	if (body == NULL || strlen(body) == 0) {
		return -1;
	}

	pCopyPos  = (char *)body;
	pStartPos = (char *)body;
	pEndPos   = (char *)body;
	while (1) {
		// find flag.
		while (*pEndPos != 0x0D || *(pEndPos+1) != 0x0A) {
			pEndPos++;
			if (*pEndPos == 0) {
				// error.
				return -1;
			}
		}
		// charge data.
		memset(szSectionLen, 0, sizeof(szSectionLen));
		strcpy(szSectionLen, "0x");
		memcpy(szSectionLen + 2, pStartPos, pEndPos - pStartPos);
    nSectionLen = strtol(szSectionLen, NULL, 16);
		if (nSectionLen == 0)
      break;
    pStartPos = pEndPos + 2;
    pEndPos   = pStartPos + nSectionLen;
    memcpy(pCopyPos, pStartPos, pEndPos - pStartPos);
    pCopyPos += pEndPos - pStartPos;
    pEndPos  += 2;
    pStartPos = pEndPos;
	}
	*pCopyPos = 0;
	return 0;
}


// 返回读取到的数据长度。
int socket_receive(int sockFd, int length, char *data)
{
	int ret, readLength, contentLength = -1, zipFlag = -1, chunkedFlag = -1, finishFlag = -1;
	struct timeval tv;
	fd_set rfds;
	char *p, *p1 = NULL, *body = NULL, chLength[16];

	readLength = 0;
	p = (char *)data;
	*p = 0;

	while (1) {
		FD_ZERO(&rfds);
		FD_SET(sockFd, &rfds);

		tv.tv_sec  = timeout_time > 6 ? timeout_time : 6;
		tv.tv_usec = 0;

		ret = select(sockFd + 1, &rfds, NULL, NULL, &tv);
		if (ret <= 0) {
			/* time out or error */
			finishFlag = 0;
			goto analyze_data;
		}

		if (FD_ISSET(sockFd, &rfds)) {
			ret = recv(sockFd, p, length-readLength, 0);
			if (ret < 0) {
				/* read error */
				dlg->ShowMessage("Read socket error.");
				finishFlag = 0;
				goto analyze_data;
			}
			if (ret == 0) {
				/* time out, the data should be error. */
				finishFlag = 0;
				goto analyze_data;
			}
			readLength += ret;
			p += ret;
			*p = 0;

analyze_data:
			if (readLength >= length) {
				/* buffer is full, return */
				dlg->ShowMessage("The buffer is not enough for read the data.");
				return -1;
			}
			// get the length of web site.
			if (contentLength == -1 && chunkedFlag == -1) {
				p1 = strstr(data, "Transfer-Encoding: chunked");
				if (p1 != NULL) {
					chunkedFlag = 0;
				} else
				if ((p1 = strstr(data, "Transfer-Encoding:  chunked")) != NULL) {
					chunkedFlag = 0;
				} else {
					p1 = strstr(data, "Content-Length: ");
					if (p1 != NULL) {
						p1 += strlen("Content-Length: ");
						ret = 0;
						while (*p1 != 0 && isdigit(*p1) != 0) {
							chLength[ret] = *p1;
							p1++;
							ret++;
						}
						chLength[ret] = 0;
						contentLength = atoi(chLength);
					}
				}
			}
			// check head finished flag.
			if (body == NULL) {
				// oh my god, this web site uses gzip. I'm crying.
				p1 = strstr(data, "Content-Encoding: gzip");
				if (p1 != NULL) {
					zipFlag = 0;
				}
				p1 = strstr(data, "Content-encoding: gzip");
				if (p1 != NULL) {
					zipFlag = 0;
				}
				p1 = strstr(data, "\r\n\r\n");
				if (p1 == NULL) {
					p1 = strstr(data, "\n\n");
					if (p1 == NULL) {
						// not find the flag, continue read data.
					} else {
						p1 += 2;
						body = p1;
					}
				} else {
					p1 += 4;
					body = p1;
				}
			}
			if (body != NULL && chunkedFlag == 0) {
				// oh, my god. this web page use chunked.
				if (*(p-1) == 0x0A && *(p-2) == 0x0D && *(p-3) == 0x0A && *(p-4) == 0x0D &&
					  *(p-5) == 0x30 && *(p-6) == 0x0A && *(p-7) == 0x0D) {
					// find the end of packed.
					finishFlag = 0;
				}
			}
			// check length.
			if (finishFlag == 0 || (contentLength > 0 && (readLength - (body-(char *)data)) >= contentLength)) {
				// finished. write the data to the file and unzip it.
				if (zipFlag == 0) {
					// it's a zip.
					if (chunkedFlag == 0) {
						ret = translate_zipdata_chunked(body, readLength - (body-(char *)data), data, length);
					} else {
						//write_data_to_file(data, readLength);
						ret = translate_zipdata(body, readLength - (body-(char *)data), data, length);
					}
					if (ret < 0) {
						//write_data_to_file(p1, contentLength);
						return 0;
					} else {
						return ret;
					}
				} else {
					// it's a normal web page.
					socket_get_cookie(data, body);
					if (chunkedFlag == 0) {
						socket_analyze_chunked(body);
					}
					return readLength;
				}
			}
		}
		/* oh my god, there are some web page include </html> in the content of html, so we can't to do so.
		if (strstr(data, "</html>") || strstr(data, "</HTML>")) {
			// find the end flag of html.
			socket_get_cookie(data, body);
			return readLength;
		}
		*/
	}
}


int socket_send(int sockFd, int length, void *data)
{
	int ret;
	int sendLength;
	char *p;

	sendLength = length;
	p = (char *)data;

	while (sendLength > 0) {
		ret = send(sockFd, p, sendLength, 0);
		if (ret <= 0) {
			/* error on send */
			return -1;
		}
		sendLength -= ret;
		p += ret;
	}
	return 0;
}













int clear_last_char(char *info)
{
	if (info == NULL || strlen(info) == 0) {
		return -1;
	}
	info[strlen(info)-1] = 0;
	return 0;
}

// write the data to file. run gunzip, read new buffer.
int translate_zipdata(const char *buffer, int length, char *data, int len)
{
	FILE *fp;
	char filename[256], cmd[512];
	long fileSize;
  pid_t pid;
  int   ret;

  pid = getpid();
	// write gzip data to file.
	sprintf(filename, "temp%d.gz", pid);
	fp = fopen(filename, "wb+");
	if (fp == NULL)
		return -1;
	fwrite(buffer, length, 1, fp);
	fclose(fp);
	// unzip.
	sprintf(cmd, "gunzip %s", filename);
	ret = system(cmd);
	// read the unziped data to buffer.
	sprintf(filename, "temp%d", pid);
	fp = fopen(filename, "rb");
	if (fp == NULL)
		return -1;
	fseek(fp, 0, 2);
	fileSize = ftell(fp);
	fseek(fp, 0, 0);
	if (fileSize < len) {
		ret = fread(data, fileSize, 1, fp);
		data[fileSize] = 0;
	}
	fclose(fp);
	unlink(filename);
	if (fileSize < len) {
		return fileSize;
	} else {
		dlg->ShowMessage("The size of buffer is not enough for read the file(b:%d,f:%ld)!", len, fileSize);
		return -1;
	}
}

// write the data to file. run gunzip, read new buffer.
int translate_zipdata_chunked(const char *buffer, int length, char *data, int len)
{
	FILE *fp;
	char filename[128], cmd[256];
	long fileSize;
	char szSectionLen[64], *pEndPos = NULL, *pStartPos = NULL;
	int  ret, ibuf, nSectionLen;
  pid_t pid;

  pid = getpid();
	// write gzip data to file.
	sprintf(filename, "temp%d.gz", pid);
	fp = fopen(filename, "wb+");
	if (fp == NULL)
		return -1;
	ibuf  		= 0;
	pStartPos = (char *)buffer;
	pEndPos   = (char *)buffer;
	while (ibuf < length) {
		// find flag.
		while (*pEndPos != 0x0D || *(pEndPos+1) != 0x0A) {
			pEndPos++;
			ibuf++;
			if (ibuf == length) {
				//error.
				fclose(fp);
				return -1;
			}
		}
		// charge data.
		memset(szSectionLen, 0, sizeof(szSectionLen));
		strcpy(szSectionLen, "0x");
		memcpy(szSectionLen + 2, pStartPos, pEndPos - pStartPos);
    nSectionLen = strtol(szSectionLen, NULL, 16);
		if (nSectionLen == 0)
      break;
    pStartPos = pEndPos + 2;
    pEndPos   = pStartPos + nSectionLen;
    ibuf 		 += 2;		// skip "\r\n"
    ibuf		 += nSectionLen;
    fwrite(pStartPos, pEndPos - pStartPos, 1, fp);
    pEndPos  += 2;
    pStartPos = pEndPos;
    ibuf 		 += 2;		// skip "\r\n"
	}
	fclose(fp);
	// unzip.
	sprintf(cmd, "gunzip %s", filename);
	ret = system(cmd);
	// read the unziped data to buffer.
	sprintf(filename, "temp%d", pid);
	fp = fopen(filename, "rb");
	if (fp == NULL)
		return -1;
	fseek(fp, 0, 2);
	fileSize = ftell(fp);
	fseek(fp, 0, 0);
	if (fileSize < len) {
		ret = fread(data, fileSize, 1, fp);
		data[fileSize] = 0;
	}
	fclose(fp);
	unlink(filename);
	if (fileSize < len) {
		return fileSize;
	} else {
		dlg->ShowMessage("The size of buffer is not enough for read the file(b:%d,f:%ld)!", len, fileSize);
		return -1;
	}
}

